/**
 *  \file
 *  \author Ricardo Cayssials
 *  \brief GeMRTOS dining philosophers example
 *  \version 0.9a
 *  \date      2015-2023
 *  \bug  None known
 *  \copyright Ricardo Cayssials  (rcayssials@gemrtos.com)
 *  \copyright All rights reserved.
 *  \copyright Copyright (c) 2013-2020
 *  \warning This product was produced by Ricardo Cayssials in the hope that it
 *  will be useful, but WITHOUT ANY WARRANTY; without even the implied
 *  warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *  \warning DISCLAIMER OF WARRANTY. THIS PRODUCT IS LICENSED "AS IS." YOU BEAR THE RISK OF USING IT.
 *  \warning RICARDO CAYSSIALS GIVES NO EXPRESS WARRANTIES, GUARANTEES, OR CONDITIONS.
 *  TO THE EXTENT PERMITTED UNDER APPLICABLE LAWS, RICARDO CAYSSIALS EXCLUDES ALL IMPLIED WARRANTIES,
 *  INCLUDING MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, AND NON-INFRINGEMENT.
 */

/******************************************************************************
*                                                                             *
* License Agreement                                                           *
* Copyright (c) Ricardo L. Cayssials                                          *
* All rights reserved.                                                        *
*                                                                             *
******************************************************************************/

/* Include the GeMRTOS library */
#include <gemrtos.h>


// ####################################################################################
// Defining system parameters
#define NUMBER_PHILOSOPHERS 30

// ####################################################################################
// Global structure for task information storage

t_semaphore_resource *leds_sem;

/* Semaphore to exclude forks variable access */
t_semaphore_resource *forks[NUMBER_PHILOSOPHERS];
#define left_fork(i)  (i)
#define right_fork(i) ((i + 1) % NUMBER_PHILOSOPHERS)

/* leds variable retains the current value written to GeMRTOS outputs */
volatile unsigned int leds;

/* Producer consumer queue for jtag_uart output*/
G_RCB  *pqueue_out;

/* Producer consumer queue for jtag_uart input*/
G_RCB  *pqueue_in;

/* Trigger for jtag output server */
G_RCB *jtag_uart_0_out;

/* Trigger for jtag input server */
G_RCB *jtag_uart_0_in;

/******************************************************************/
/* Signal function when task is aborted (the deadline is missed)  */
/******************************************************************/
void sig_aborted_task_generic(int pdata)
{
    gu_printf("\nTASK %u ABORTED\n", (unsigned int) pdata);
}

/******************************************************************/
/* generic task function                                          */
/******************************************************************/
void task_philosopher(void* pdata)
{

	unsigned int i = (unsigned int) pdata;

	/* get the pointer to the TCB of the current task */
	GS_TCB *ptcb = gu_PCB_GetCurrentTCB();

    // 13-09-2024 PRINT_ASSERT((i < NUMBER_PHILOSOPHERS),"ERROR In i, i=%u, arg=%u, %s\n", i, (unsigned int) ptcb->TCB_TaskArg, ptcb->TCB_description);          // !!!!D

    // 13-09-2024 PRINT_ASSERT((gkm_RCB_IsValid(forks[left_fork(i)]) == G_TRUE),"ERROR Left fork not valid, value = %p, i=%u\n", (void *) forks[left_fork(i)], i);      // !!!!D
    // 13-09-2024 PRINT_ASSERT((gkm_RCB_IsValid(forks[right_fork(i)]) == G_TRUE),"ERROR Right fork not valid, value = %p, i=%u\n", (void *) forks[right_fork(i)], i);    // !!!!D
	
	/* get the semaphores in order */
	if (( i & 1) == 0) {  // philosopher is even
		gu_sem_wait(forks[left_fork(i)], (int) G_TRUE);
		gu_sem_wait(forks[right_fork(i)], (int) G_TRUE);
	} else {              // philosopher is odd
		gu_sem_wait(forks[right_fork(i)], (int) G_TRUE);
		gu_sem_wait(forks[left_fork(i)], (int) G_TRUE);
	}

	/* change the leds value             */
    leds = leds + 1; // ^ (unsigned int) ptcb; // xor with ptcb
	/* write the value in leds variable to the GeMRTOS output port  */
	gu_write_outputs(leds);
 
	gu_printf("%s is eating in processor %u\n", ptcb->TCB_description, (unsigned int) GRTOS_CMD_PRC_ID);
	if (i > NUMBER_PHILOSOPHERS / 2)
		gu_queue_printf(pqueue_out,"%s is eating in processor %d\n", ptcb->TCB_description, (int) GRTOS_CMD_PRC_ID);	

	/* release the semaphore in order */
	gu_sem_post((G_RCB *) forks[left_fork(i)]);
	gu_sem_post((G_RCB *) forks[right_fork(i)]);	
}


struct jtag_uart_info {
	unsigned int trigger_out;
	unsigned int trigger_in;
	G_RCB *pqueue_out;
	G_RCB *pqueue_in;
	unsigned int buffer_length_out;
	unsigned int buffer_length_in;
	unsigned int jtag_base;
};

#ifdef JTAG_UART_0_BASE
void *GRTOS_IRQ_READ_SERVER_ENABLE(void * pdata)
{
#ifdef __niosX_arch__       // !!!!D
	PRINT_ASSERT((GRTOS_CMD_PRC_ID == GRTOS_MTX_PRC_GRANTED), "ERROR MTX not proc=%d, mtx=%d\n",(int) GRTOS_CMD_PRC_ID, (int) GRTOS_MTX_PRC_GRANTED);

	// Enable interrupt in JTAG UART component
	unsigned int status = IORD_ALTERA_AVALON_JTAG_UART_CONTROL(JTAG_UART_0_BASE);
	IOWR_ALTERA_AVALON_JTAG_UART_CONTROL(JTAG_UART_0_BASE, status | ALTERA_AVALON_JTAG_UART_CONTROL_RE_MSK);

	// 13-09-2024 // Dummy read to take effect                                    // !!!!D
	// 13-09-2024 IORD_ALTERA_AVALON_JTAG_UART_CONTROL(JTAG_UART_0_BASE);         // !!!!D
#endif                     // !!!!D  __niosX_arch__
	return ((void *) 0);
}

void *GRTOS_IRQ_READ_SERVER_DISABLE(void *pdata)
{
#ifdef __niosX_arch__       // !!!!D
   PRINT_ASSERT((GRTOS_CMD_PRC_ID == GRTOS_MTX_PRC_GRANTED), "ERROR MTX not proc=%d, mtx=%d\n",(int) GRTOS_CMD_PRC_ID, (int) GRTOS_MTX_PRC_GRANTED);

	// Disable interrupt in JTAG UART component
	unsigned int status = IORD_ALTERA_AVALON_JTAG_UART_CONTROL(JTAG_UART_0_BASE);
	IOWR_ALTERA_AVALON_JTAG_UART_CONTROL(JTAG_UART_0_BASE, status & ~(ALTERA_AVALON_JTAG_UART_CONTROL_RE_MSK));

	// 13-09-2024 // Dummy read to take effect                                     // !!!!D
	// 13-09-2024 IORD_ALTERA_AVALON_JTAG_UART_CONTROL(JTAG_UART_0_BASE);          // !!!!D
#endif                     // !!!!D  __niosX_arch__
    return ((void *) 0);
}

void GRTOS_IRQ_READ_SERVER(void *pdata)
{

	struct jtag_uart_info * jtag_uart_0_info = (struct jtag_uart_info *) pdata;

	unsigned int index_msg = 0;
	unsigned int uart_data;

	char *buffer_msg = (char *) malloc(jtag_uart_0_info->buffer_length_in);	
		PRINT_ASSERT(((unsigned int) buffer_msg % 4 == 0), "ERROR memory is not alligned");
		PRINT_ASSERT(((unsigned int) buffer_msg != 0), "ERROR no memory available");

    /* Get the pointer to the TCB of the current task */
    GS_TCB *ptcb = gu_PCB_GetCurrentTCB();	
	
    /* Define this task as an ISR associated to trigger_out jtag_uart_0_in */
    gu_trigger_register_task(ptcb, (G_INT32) jtag_uart_0_info->trigger_in);
	
	while (1) {
		
		/* wait for trigger to read the jtag buffer */
		gu_trigger_wait();

#ifdef __niosX_arch__       // !!!!D

		while ((uart_data = IORD_ALTERA_AVALON_JTAG_UART_DATA(JTAG_UART_0_BASE)) & ALTERA_AVALON_JTAG_UART_DATA_RVALID_MSK) {

			char data = (char) (uart_data & ALTERA_AVALON_JTAG_UART_DATA_DATA_MSK) >> ALTERA_AVALON_JTAG_UART_DATA_DATA_OFST;

			buffer_msg[index_msg] = data;
			index_msg++;
			// GU_FPRINTF(stderr, "index_msg= %d, character = %c\n", index_msg, data);  // !!!!D
			
			if ((data == '\r') || (data == '\n') || (index_msg >= jtag_uart_0_info->buffer_length_in - 2)) {
				buffer_msg[index_msg] = '\0'; // end of string
				/* Produce the message received to the queue_in */
				gu_fprintf("\nINPUT FROM JTAG-UART = %s\n", buffer_msg);
				gu_queue_printf(jtag_uart_0_info->pqueue_in,"%s", buffer_msg);
				index_msg = 0;
				break;
			}
		}
#endif                     // !!!!D  __niosX_arch__
	}

}

void *GRTOS_IRQ_WRITE_SERVER_ENABLE(void * pdata)
{
#ifdef __niosX_arch__       // !!!!D
	PRINT_ASSERT((GRTOS_CMD_PRC_ID == GRTOS_MTX_PRC_GRANTED), "ERROR MTX not proc=%d, mtx=%d\n",(int) GRTOS_CMD_PRC_ID, (int) GRTOS_MTX_PRC_GRANTED);


	// Enable interrupt in JTAG UART component
	unsigned int status = IORD_ALTERA_AVALON_JTAG_UART_CONTROL(JTAG_UART_0_BASE);
	IOWR_ALTERA_AVALON_JTAG_UART_CONTROL(JTAG_UART_0_BASE, status | ALTERA_AVALON_JTAG_UART_CONTROL_WE_MSK);

	// 13-09-2024 // Dummy read to take effect                                    // !!!!D
	// 13-09-2024 IORD_ALTERA_AVALON_JTAG_UART_CONTROL(JTAG_UART_0_BASE);         // !!!!D
#endif                     // !!!!D  __niosX_arch__
    return ((void *) 0);
}

void *GRTOS_IRQ_WRITE_SERVER_DISABLE(void *pdata)
{
#ifdef __niosX_arch__       // !!!!D
	PRINT_ASSERT((GRTOS_CMD_PRC_ID == GRTOS_MTX_PRC_GRANTED), "ERROR MTX not proc=%d, mtx=%d\n",(int) GRTOS_CMD_PRC_ID, (int) GRTOS_MTX_PRC_GRANTED);

	// Disable interrupt in JTAG UART component
	unsigned int status = IORD_ALTERA_AVALON_JTAG_UART_CONTROL(JTAG_UART_0_BASE);
	IOWR_ALTERA_AVALON_JTAG_UART_CONTROL(JTAG_UART_0_BASE, status & ~(ALTERA_AVALON_JTAG_UART_CONTROL_WE_MSK));

	// 13-09-2024 // Dummy read to take effect                                     // !!!!D
	// 13-09-2024 IORD_ALTERA_AVALON_JTAG_UART_CONTROL(JTAG_UART_0_BASE);          // !!!!D
#endif                     // !!!!D  __niosX_arch__
    return ((void *) 0);
}


void GRTOS_IRQ_WRITE_SERVER(void *pdata)
{

	struct jtag_uart_info * jtag_uart_0_info = (struct jtag_uart_info *) pdata;
	
	/* Get a buffer to store the messaje to send to the jtag-uart */
    char *buffer_msg = (char *) malloc(jtag_uart_0_info->buffer_length_out);
		PRINT_ASSERT(((unsigned int) buffer_msg % 4 == 0), "ERROR memory is not alligned");
		PRINT_ASSERT(((unsigned int) buffer_msg != 0), "ERROR no memory available");
    
    int size;
    unsigned int index_msg;
	
    /* Get the pointer to the TCB of the current task */
    GS_TCB *ptcb = gu_PCB_GetCurrentTCB();

    /* Define this task as an ISR associated to trigger_out jtag_uart_0_out */
    gu_trigger_register_task(ptcb, (G_INT32) jtag_uart_0_info->trigger_out);

    while (1) {
		/* receive next message from producer/consumer queue */
        size = gu_queue_receive(jtag_uart_0_info->pqueue_out, (void *) buffer_msg, jtag_uart_0_info->buffer_length_out);
		
		/* Iterate for each character of the message, but the ending '\0' */
        index_msg = 0;
        while (index_msg < size-1)
        {
#ifdef __niosX_arch__       // !!!!D
            // int space_fifo;
			/* Get the available space in jtag uart buffer */
			int space_fifo = (((G_INT32) IORD_ALTERA_AVALON_JTAG_UART_CONTROL(jtag_uart_0_info->jtag_base)) & ((G_INT32) ALTERA_AVALON_JTAG_UART_CONTROL_WSPACE_MSK )) >> ALTERA_AVALON_JTAG_UART_CONTROL_WSPACE_OFST;

			// if (space_fifo > size) space_fifo = size - 2;
			/* Send as many characters up to the end of message or as possible to jtag uart */
			while (space_fifo > 0 && index_msg < size-1)
			{
				PRINT_ASSERT((buffer_msg[index_msg] != '\0'), "Error char zero=%d", (int) buffer_msg[index_msg]);

				/* write next message character to jtag uart output fifo */
				IOWR_ALTERA_AVALON_JTAG_UART_DATA(jtag_uart_0_info->jtag_base, buffer_msg[index_msg]);

				/* Update iteration variables */
				index_msg++;
				space_fifo--;
			}
			/* There exist more character to send, so wait to next trigger_out from jtag device */
            if (index_msg < size-1) {
                gu_trigger_wait();
            }
#endif                     // !!!!D  __niosX_arch__
        }
    };
}


/* This function is the ISR for the jtag uart named "jtag_uart_0_out" */
void jtag_irq_manager(int pdata)
{
#ifdef __niosX_arch__       // !!!!D

	/* Read the state of interrupt flags from the JTAG UART status register */
    volatile G_INT32 uart_ctrl_register = (G_INT32) IORD_ALTERA_AVALON_JTAG_UART_CONTROL(JTAG_UART_0_BASE);
    
	/* Get the state of the write interrupt (output buffer empty) */
    volatile G_INT32 write_interrupt = (uart_ctrl_register  & ((G_INT32) ALTERA_AVALON_JTAG_UART_CONTROL_WI_MSK )) >> ALTERA_AVALON_JTAG_UART_CONTROL_WI_OFST;
	
    /* Get the state of the read interrupt (input buffer full) */
    volatile G_INT32 read_interrupt = (uart_ctrl_register  & ((G_INT32) ALTERA_AVALON_JTAG_UART_CONTROL_RI_MSK )) >> ALTERA_AVALON_JTAG_UART_CONTROL_RI_OFST;    

	/* If write interrupt is on, the release the output server trigger_out */
    if ((write_interrupt == (G_INT32) 1) && (jtag_uart_0_out->trigger.TRG_Enabled == G_TRUE)) {
		PRINT_ASSERT((jtag_uart_0_out->trigger.TRG_Enabled == G_TRUE),"ERROR trigger write is not enabled\n");
		gu_trigger_release ((int) jtag_uart_0_out);
	}
	
	/* If write interrupt is on, the release the output server trigger_in */
    if ((read_interrupt == (G_INT32) 1) && (jtag_uart_0_in->trigger.TRG_Enabled == G_TRUE)) {
		PRINT_ASSERT((jtag_uart_0_in->trigger.TRG_Enabled == G_TRUE),"ERROR trigger read is not enabled\n");
		gu_trigger_release ((int) jtag_uart_0_in);
	}
	PRINT_ASSERT(((write_interrupt == (G_INT32) 1) || (read_interrupt == (G_INT32) 1)),"ERROR not interrupt triggering");
#endif                     // !!!!D  __niosX_arch__
}
#endif

int main(void)
{
    /* ################################################### */
    /* main variable definition                            */
	/* ################################################### */
    GS_TCB *ptcb;

    /* Create a semaphore for LEDS */
    leds_sem = gu_sem_create((int) 1);
	
	/* Create semaphores for forks */
	for (int i = 0; i < NUMBER_PHILOSOPHERS; i++) {
		forks[i] = gu_sem_create((int) 1);
	}		
	
	// Create the philosopher
	for (unsigned int i = 0; i < NUMBER_PHILOSOPHERS; i++)
	{
		/* Create a new Task */
		ptcb = gu_GetTask((void *) task_philosopher,    ///< Pointer to the task code
						  (void *) i,                   ///< Pointer to the argument of first call
						  "task %d", i);
									   
		if (ptcb != (void *) 0) { /* Task was created successfully */
			/* Set the task as periodic */
			gu_SetTaskType(ptcb, G_TCBType_PERIODIC);
			
			/* Set the task priority equal to the task index plus one */
			gu_SetTaskReadyPriority(ptcb, (G_INT64) i+1);
			gu_SetTaskRunPriority(ptcb,(G_INT64) i+1);

			/* Set the period equal to the task index plus one in seconds */
			gu_SetTaskPeriod(ptcb, 0, 0, 1, (i+1) * 100);
			
			/* Start execution of task with an offset equal to five seconds */
			gu_StartTaskwithOffset(ptcb, 0, 0, 5, 0);
			gu_signal_create(G_SCBType_TCB_ABORTED, 0, (void *) ptcb, (void *) sig_aborted_task_generic, (void *) i);  /// Signal when task is aborted because of deadline
		}
		else {
			gu_printf("Error when creating task\n");
			while(1);
		}
	}

#ifdef JTAG_UART_0_BASE

    /* Create a message queue for JTAG-UART jtag_uart_0 output */
    pqueue_out = gu_queue_create();	
	
    /* Create a message queue for JTAG-UART jtag_uart_0 input */
    pqueue_in = gu_queue_create();

	/* Create a trigger_out for JTAG UART writting */
	/* -1 because it is not associated with hardware interrupt */
	jtag_uart_0_out = gu_trigger_create(-1);
	PRINT_ASSERT((gkm_RCB_IsValid(jtag_uart_0_out) == G_TRUE),"ERROR ECB not valid\n");
	
	/* Create a trigger_in for JTAG UART reading */
	/* -1 because it is not associated with hardware interrupt */
	jtag_uart_0_in = gu_trigger_create(-1);
	PRINT_ASSERT((gkm_RCB_IsValid(jtag_uart_0_in) == G_TRUE),"ERROR ECB not valid\n");

	/* create a jtag_uart_info structure to store the information for output server consumer */
	struct jtag_uart_info *jtag_uart_0_info = (struct jtag_uart_info *) malloc(sizeof(struct jtag_uart_info));
	
	/* set the parameter for jtag output server configuration. 
	   The pointer to this structure is pass as reference to the task */
	jtag_uart_0_info->pqueue_out = pqueue_out;
	jtag_uart_0_info->pqueue_in = pqueue_in;
	jtag_uart_0_info->trigger_out = (unsigned int) jtag_uart_0_out;
	jtag_uart_0_info->trigger_in = (unsigned int) jtag_uart_0_in;
	jtag_uart_0_info->buffer_length_in = 100;
	jtag_uart_0_info->buffer_length_out = 100;
	jtag_uart_0_info->jtag_base = JTAG_UART_0_BASE;

	/* Creatse task for jtag output server */
	ptcb = gu_GetTask((void *) GRTOS_IRQ_WRITE_SERVER,
					   (void *) jtag_uart_0_info,
					   "GRTOS_IRQ_WRITE_SERVER");
	if (ptcb != (void *) 0)
	{
		gu_SetTaskType(ptcb, G_TCBType_UCOS);

		// gu_SetTaskList(ptcb, (struct gs_lcb *) pedf_list);
		gu_SetTaskReadyPriority(ptcb, (G_INT64) 0);
		gu_SetTaskRunPriority(ptcb, (G_INT64) 0);

		gu_SetTaskPeriod(ptcb, 0, 0,  1, 1);

		/* Start execution of task with an offset */
		gu_StartTaskwithOffset(ptcb, 0, 0, 5, 0);
	}
	/* suscribe task as consumer to the consumer/producer queue */
	gu_queue_subscribe(ptcb, pqueue_out);
	
	/* Set the jtag output server hooks */
	gu_trigger_enable_hook((int) jtag_uart_0_out, (void *) GRTOS_IRQ_WRITE_SERVER_ENABLE, (void *) NULL);
	gu_trigger_disable_hook((int) jtag_uart_0_out, (void *) GRTOS_IRQ_WRITE_SERVER_DISABLE, (void *) NULL);

	/* Creatse task for jtag output server */
	ptcb = gu_GetTask((void *) GRTOS_IRQ_READ_SERVER,
					   (void *) jtag_uart_0_info,
					   "GRTOS_IRQ_READ_SERVER");
	if (ptcb != (void *) 0)
	{
		gu_SetTaskType(ptcb, G_TCBType_UCOS);
	
		// gu_SetTaskList(ptcb, (struct gs_lcb *) pedf_list);
		gu_SetTaskReadyPriority(ptcb, (G_INT64) 0);
		gu_SetTaskRunPriority(ptcb, (G_INT64) 0);
	
		gu_SetTaskPeriod(ptcb, 0, 0,  1, 1);
		
		/* Start execution of task with an offset */
		gu_StartTaskwithOffset(ptcb, 0, 0, 5, 0);
	}
	
	/* Set the jtag output server hooks */
	gu_trigger_enable_hook((int) jtag_uart_0_in, (void *) GRTOS_IRQ_READ_SERVER_ENABLE, (void *) NULL);
	gu_trigger_disable_hook((int) jtag_uart_0_in, (void *) GRTOS_IRQ_READ_SERVER_DISABLE, (void *) NULL);	
	// 22-03-2024 ####################################	

	// #################################
	/* Create a ISR task for jtag uart hardware interrupt */
	ptcb = gu_GetTask((void *) jtag_irq_manager,
					   (void *) 5, 
					   "jtag_irq_manager");
					   
	PRINT_ASSERT((gkm_TCB_IsValid(ptcb) == G_TRUE),"ERROR TCB is not valid\n");
	if (ptcb != (void *) 0)
	{
		gu_SetTaskType(ptcb, G_TCBType_UCOS);

		// gu_SetTaskList(ptcb, (struct gs_lcb *) pedf_list);
		
		gu_SetTaskReadyPriority(ptcb, (G_INT64) 0);
		gu_SetTaskRunPriority(ptcb, (G_INT64) 0);

		gu_SetTaskPeriod(ptcb, 0, 0,  1, 1);
	}        
	
	/* Associate the task to be triggered with the jtag uart interrupt */
	gu_trigger_register_task(ptcb, (int) JTAG_UART_0_IRQ);
	gu_trigger_enable((int) JTAG_UART_0_IRQ);
#endif
	
    return (0);
}

